/*
 * FreeSWITCH Modular Media Switching Software Library / Soft-Switch Application
 * Copyright (C) 2005-2014, Anthony Minessale II <anthm@freeswitch.org>
 *
 * Version: MPL 1.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is FreeSWITCH Modular Media Switching Software Library / Soft-Switch Application
 *
 * The Initial Developer of the Original Code is
 * Piotr Gregor <piotr@freeswitch.com> <piotr@signalwire.com>
 *
 * Portions created by the Initial Developer are Copyright (C)
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *
 * Piotr Gregor <piotr@freeswitch.org> <piotr@signalwire.com>
 *
 * mod_red.c -- The redundancy encoder/decoder (RFC 2198, RFC 5109)
 *
 */


#ifndef ULP_ULP_H
#define ULP_ULP_H

#include <stdint.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#ifndef WIN32
#include <arpa/inet.h>
#endif

#ifdef __cplusplus
extern "C" {
#endif

typedef enum ulp_direction {
    ulp_direction_sender = 1,
    ulp_direction_receiver = 2
} ulp_direction_t;

/**
 * Maximum number of protection levels that fit into 48 bitmask, assuming they are 2 packets
 * aligned and requiring that if packet contains ULP FEC of level k then it must contain all
 * protection levels up to the k.
 *
 * This implies that maximum number of levels (groups) is 5. In such maximum case, the size
 * of group L0 is G(0) = 2, and all group sizes are: G(k) = 2 ^ (k + 1). All numbers
 * of protection levels up to and including ULP_LVL_N are supported.
 *
 * Reference: RFC 5109: 7.4.  FEC Level Header for FEC Packets.
 */
#define ULP_LVL_N 5

/**
 * Max number of FECs used to recover RTP packet on single level of protection.
 */
#define ULP_FEC_MAX_N 40

/* Maximum number of packets protected by single level. Limited by 16 bits or 48 bits size
 * of the mask in FEC Level Header. */
#define ULP_LVL_GROUP_SIZE_MAX 48

#define ULP_LVL_LEN_MAX 16384
#define ULP_RTP_LEN_MAX ULP_LVL_LEN_MAX

#define ULP_PT_RED 101
#define ULP_PT_ULPFEC 127
#define ULP_MAX_FEC_IN_SCHEME 12

#define ULP_MASK16_TABLE_SIZE 78
#define ULP_MASK16_RAND_LOSS_TABLE_SIZE 12
#define ULP_MAX_MASK_SETS 10 /* 5% steps up to 50% */
/*
 * ulp_hdr_t represents an RTP header.  The bit-fields in
 * this structure should be declared "unsigned int" instead of
 * "unsigned char", but doing so causes the MS compiler to not
 * fully pack the bit fields.
 *
 * In this implementation, an ulp_hdr_t is assumed to be 32-bit aligned
 *
 * (note that this definition follows that of RFC 1889 Appendix A, but
 * is not identical)
 */

#ifndef WORDS_BIGENDIAN

typedef struct PACKED {
    unsigned cc : 4;		/* CSRC count             */
    unsigned x : 1;			/* header extension flag  */
    unsigned p : 1;			/* padding flag           */

    unsigned v : 2;			/* protocol version       */

    unsigned pt : 7;		/* payload type           */
    unsigned m : 1;			/* marker bit             */
    uint16_t seq;			/* sequence number        */
    uint32_t ts;			/* timestamp              */
    uint32_t ssrc;			/* synchronization source */
} ulp_rtp_hdr_t;

typedef struct PACKED {
    unsigned cc_r : 4;		/* CC recovery		*/
    unsigned x_r : 1;		/* X recovery		*/
    unsigned p_r : 1;		/* P recovery		*/

    unsigned l : 1;			/* long-mask		*/
    unsigned e : 1;			/* extension		*/

    unsigned pt_r : 7;		/* PT recovery		*/
    unsigned m_r : 1;		/* M recovery		*/
    uint16_t snb;			/* seq base			*/
    uint32_t ts_r;			/* TS recovery		*/
    uint16_t len_r;			/* Length recovery	*/
} ulp_fec_hdr_t;

typedef struct PACKED {
	unsigned pt : 7;		/* RTP payload type for this encoding block. For primary encoding this should be same as that of the RTP media packet. */
	unsigned f : 1;			/* Flag 1 means redundant encoding, 0 - primary.	*/
	uint8_t ts_off_len[3];	/* Timestamp offset from primary and length of the payload.
							   RFC 2198 2: Redundant blocks of data will correspond to different time intervals than the primary
							   data, and hence each block of redundant encoding will require its own timestamp.
							   To reduce the number of bytes needed to carry the timestamp, it can be encoded as the difference
							   of the timestamp for the redundant encoding and the timestamp of the primary. */
							/* Length in bytes of the corresponding data block excluding this header. */
} ulp_red_encoding_block_hdr_t;

/**
 * TS - 14 bits, LEN - 10 bits
 */
#define ulp_red_encoding_block_get_ts(ts_off_len) ((ts_off_len[0] << 6) | (((ts_off_len[1]) >> 2) & 0x3Fu))
#define ulp_red_encoding_block_get_len(ts_off_len) (ts_off_len[2] | ((ts_off_len[1] & 0x3u) << 8))

/**
 * Set TS and LEN fields of Encoding Block Header to @ts and @len.
 * @ts_off_len - must point to 3 octets of TS and LEN in the Encoding Block Header
 * @ts - offset of the payload from the RTP media
 * @len - length of the payload following Encoding Block Header 
 */
SWITCH_DECLARE(void) ulp_red_encoding_block_set_ts_off_len(uint8_t *ts_off_len, uint16_t ts, uint16_t len);

#else /*  BIG_ENDIAN */

typedef struct __attribute__((__packed__)) {
    unsigned char v : 2;		/* protocol version       */
    unsigned char p : 1;		/* padding flag           */
    unsigned char x : 1;		/* header extension flag  */
    unsigned char cc : 4;		/* CSRC count             */
    unsigned char m : 1;		/* marker bit             */
    unsigned char pt : 7;		/* payload type           */
    uint16_t seq;				/* sequence number        */
    uint32_t ts;				/* timestamp              */
    uint32_t ssrc;				/* synchronization source */
} ulp_rtp_hdr_t;

/**
 * The P recovery field, the X recovery field, the CC recovery field, the M recovery field,
 * and the PT recovery field are obtained via the protection operation applied
 * to the corresponding P, X, CC, M, and PT values from the RTP header of the media
 * packets associated with the FEC packet.
 */
typedef struct __attribute__((__packed__)) {
    unsigned char e : 1;		/* extension. The E bit is the extension flag reserved to indicate any future extension to this specification. It SHALL be set to 0, and SHOULD be ignored by the receiver.*/
    unsigned char l : 1;		/* long-mask. The L bit indicates whether the long mask is used.  When the L bit is not set, the mask is 16 bits long. When the L bit is set, the mask is then 48 bits long. */
    unsigned char p_r : 1;		/* P recovery		*/
    unsigned char x_r : 1;		/* X recovery		*/
    unsigned char cc_r : 4;		/* CC recovery		*/
    unsigned char m_r : 1;		/* M recovery		*/
    unsigned char pt_r : 7;		/* PT recovery		*/
    uint16_t snb;				/* seq base			*/
    uint32_t ts_r;				/* TS recovery		*/
    uint16_t len_r;				/* Length recovery	*/
} ulp_fec_hdr_t;

typedef struct __attribute__((__packed__)) {
    uint8_t f : 1;				/* Flag 1 means redundant encoding, 0 - primary.	*/
	uint8_t pt : 7;				/* RTP payload type for this encoding block. For primary encoding this should be same as that of the RTP media packet. */
	uint8_t ts_off_len[3];		/* Timestamp offset from primary and length of the payload.
								   RFC 2198 2: Redundant blocks of data will correspond to different time intervals than the primary
								   data, and hence each block of redundant encoding will require its own timestamp.
								   To reduce the number of bytes needed to carry the timestamp, it can be encoded as the difference
								   of the timestamp for the redundant encoding and the timestamp of the primary. */
								/* Length in bytes of the corresponding data block excluding this header. */
} ulp_red_encoding_block_hdr_t;

#endif

typedef struct {
    uint16_t profile_specific; /* profile-specific info               */
    uint16_t length;           /* number of 32-bit words in extension */
} ulp_rtp_hdr_xtnd_t;

typedef struct {
	ulp_rtp_hdr_t		hdr;
	char				b[ULP_RTP_LEN_MAX];
} ulp_rtp_pkt_t;

typedef struct {
	ulp_rtp_pkt_t		*rtp;
	uint16_t			payload_len;			/* Actual length of the payload (may indicate updated length changed by padding or truncated). */
	uint16_t			payload_original_len;	/* Original length of the RTP packet, equal to @payload_len initially, but @payload_len may be changed due to padding in ulp_rtp_cover(). */
} ulp_rtp_msg_t;


/* FEC level header is 4 or 8 octets depending on long-mask flag in FEC header. */

/* FEC level for short mask (L bit == 0 in FEC header). */
typedef struct {
	uint16_t lk;				/* protection length	*/
	uint16_t mask16;			/* protection mask. Associates media with this protection level. If bit i is set in mask to 1, then media packet with seq N + i is protected by this FEC level, N = SN Base, MSB corresponds to 0, LSB to 15 or 47. */
} ulp_fec_lvl_32_t;

/* FEC level for long mask (L bit == 1 in FEC header). */
typedef struct {
	uint16_t lk;				/* protection length	*/
	uint16_t mask16;			/* protection mask. Associates media with this protection level. If bit i is set in mask to 1, then media packet with seq N + i is protected by this FEC level, N = SN Base, MSB corresponds to 0, LSB to 15 or 47. */
	uint32_t mask_cont;			/* protection mask. Associates media with this protection level. If bit i is set in mask to 1, then media packet with seq N + i is protected by this FEC level, N = SN Base, MSB corresponds to 0, LSB to 15 or 47. */
} ulp_fec_lvl_64_t;

/**
 * Levels contain host ordered info.
 *
 * @lk - protection length
 * @mask16 - protection mask (host order). Associates media with this protection level. If bit i is set in mask to 1, then media packet with seq N + i is protected by this FEC level, N = SN Base, MSB corresponds to 0, LSB to 15 or 47.
 * @mask_cont - protection mask (host order). Associates media with this protection level. If bit i is set in mask to 1, then media packet with seq N + i is protected by this FEC level, N = SN Base, MSB corresponds to 0, LSB to 15 or 47.
 * @gk - group size of this protection level
 */
typedef struct {
	uint16_t lk;
	uint16_t mask16;
	uint32_t mask_cont;
	uint8_t gk;
} ulp_fec_lvl_64_info_t;

typedef struct {
	ulp_fec_hdr_t		hdr;	/* 10 octets. */
	ulp_fec_lvl_32_t	lvl[1];	/* Variable length depending on number of levels included and the long-mask bit. */
} ulp_fec_pkt_t;
		
typedef struct {
	ulp_rtp_hdr_t					hdr;
	ulp_red_encoding_block_hdr_t	en;	/* First of the encodings, may be Primary or Redundant encoding. */
} ulp_red_pkt_t;

typedef enum {
    ulp_status_ok = 0,					/* OK */
    ulp_status_fail = 1,				/* Not OK */
	ulp_status_recovery_partial = 2,	/* recovery succeeded partially */
	ulp_status_recovery_no_fecs = 3,	/* recovery failed due to missing FECs */ 
	ulp_status_recovery_more_rtp = 4,	/* recovery failed due to missing RTP packets */
	ulp_status_recovery_err_mem = 5		/* recovery failed due to system I/O error */
} ulp_status_t;

extern const char* ulp_status_2_str[];

typedef struct {
	ulp_rtp_msg_t	buf[ULP_LVL_GROUP_SIZE_MAX];
	uint8_t			getp;							/* Points to the oldest element. */
	uint8_t			putp;							/* Points to the available space where to write the entry.
														If getp == putp buffer is EMPTY.
														If (putp + 1) % ULP_LVL_GROUP_SIZE_MAX == getp, buffer is FULL. */
	uint8_t			getp_soft;						/* Moved with _get_soft to the next element, not freeing the previous one. */
} ulp_circ_t;	

typedef struct ulp_policy_s ulp_policy_t;

typedef struct {
	uint16_t *mask_arr;
	uint8_t media_packets;
	uint8_t fec_packets;
} mask_set_t;

/**
 * Copy requested RTP packets needed by FEC with @snb on @k level of protection
 * into @rtp array passed to this callback. Return the number of packets copied.
 * ULP owns the RTP packets and will free them.
 *
 * This callback is called serially when ULP generates FEC header for all FEC levels, k = [0, fec_n):
 * cb_rtp_material { @policy, @snb, @k } as requested by user in a call to
 *
 *		ulp_fec_pkt_t* ulp_gen_fec(ulp_policy_t *policy, uint8_t fec_n, uint16_t snb)
 *
 * The levels of protection are descibed by the policy->lvl table. Each level @k contains a bitmask
 * of packet sequences which are needed by that level. This packet mask and @snb (sequence number base)
 * completely define RTP sequences needed by FEC on that @k level.
 *
 * If the number of RTP is smaller than needed to generate FEC packet, then ulp_gen_fec() will fail.
 */
typedef uint8_t (SWITCH_THREAD_FUNC *ulp_cb_rtp_material)(ulp_policy_t *policy, uint16_t snb, uint8_t k, ulp_rtp_msg_t *rtp);

/**
 * Return copy of the RTP with sequence @seq. Set @payload_original_len to length of the packet's body.
 */
typedef ulp_rtp_pkt_t* (SWITCH_THREAD_FUNC *ulp_cb_rtp_by_seq)(ulp_policy_t *policy, uint16_t seq, uint16_t *payload_original_len);

/**
 * Return a copy of the FEC packet needed for recovery of @seq on level @k.
 */
typedef uint8_t (SWITCH_THREAD_FUNC *ulp_cb_find_fecs_for_seq_on_lvl)(ulp_policy_t *policy, uint16_t seq, uint8_t k, ulp_rtp_msg_t *fecs);

/**
 * Return copies of packets in the @rtp table. ULP owns them and calls free on them.
 */
typedef uint8_t (SWITCH_THREAD_FUNC *ulp_cb_rtp_cover)(ulp_policy_t *policy, ulp_fec_pkt_t *fec, uint16_t fec_payload_len, uint8_t k, ulp_rtp_msg_t *rtp);

struct ulp_policy_s {
	ulp_fec_lvl_64_info_t lvl[ULP_LVL_N];			/* Table of configurable levels of protection */

	uint8_t lvl_n;									/* Assigned by ulp_fec_parse: Number of protection levels found in most recent parsed FEC packet. */
	uint8_t more;									/* Assigned by ulp_fec_parse: Set to 1 if there was more levels of protection than expected (by argument passed to ulp_fec_parse). */
	uint8_t l;										/* Assigned by ulp_fec_parse: Set to 1 if FEC uses 48 bits of protection mask. */

	uint32_t ssrc;
	uint32_t flags;

	uint16_t	fec_last_len;						/* Length of FEC packet generated by last call to ulp_gen_fec(). */
	uint16_t	rtp_last_len;						/* Total length of recently recovered RTP packet generated by last call to ulp_recover(). The recovered length (restored to original) may be smaller than this. */
	uint16_t	rtp_last_recover_len;				/* Recovered length (restored to original) of recently recovered RTP packet generated by last call to ulp_recover(). */

	ulp_cb_rtp_material				cb_rtp_material;

	ulp_cb_rtp_by_seq				cb_rtp_by_seq;
	ulp_cb_find_fecs_for_seq_on_lvl	cb_find_fecs_for_seq_on_lvl;
	ulp_cb_rtp_cover				cb_rtp_cover;

	size_t   attempted_recovery_n;           /* All recovery attempts. */
	size_t   attempted_missing_fec;          /* attempts failed due to missing FEC */
	size_t   attempted_missing_media;        /* attempt failed to to missing RTP (media) */
	size_t   successful_recovery_n;          /* Successful recoveries (including partial recoveries). */
	size_t   successful_partial_recovery_n;  /* Only partially successful recoveries. */

	void *user_data;
};

SWITCH_DECLARE(ulp_status_t) ulp_policy_init(ulp_policy_t *policy,
										uint32_t ssrc,
										ulp_cb_rtp_material cb_rtp_material,
										void* user_data,
										ulp_cb_rtp_by_seq cb_rtp_by_seq,
										ulp_cb_find_fecs_for_seq_on_lvl cb_find_fecs_for_seq_on_lvl,
										ulp_cb_rtp_cover cb_rtp_cover);


SWITCH_DECLARE(void) ulp_policy_deinit(ulp_policy_t *policy);

SWITCH_DECLARE(size_t) ulp_attempted_recovery_n(ulp_policy_t *policy);
SWITCH_DECLARE(void) ulp_increment_attempted_recovery_n(ulp_policy_t *policy);

SWITCH_DECLARE(size_t) ulp_attempted_missing_fec(ulp_policy_t *policy);
SWITCH_DECLARE(void) ulp_increment_attempted_missing_fec(ulp_policy_t *policy);

SWITCH_DECLARE(size_t) ulp_attempted_missing_media(ulp_policy_t *policy);
SWITCH_DECLARE(void) ulp_increment_attempted_missing_media(ulp_policy_t *policy);

SWITCH_DECLARE(size_t) ulp_successful_recovery_n(ulp_policy_t *policy);
SWITCH_DECLARE(void) ulp_increment_successful_recovery_n(ulp_policy_t *policy);

SWITCH_DECLARE(size_t) ulp_successful_partial_recovery_n(ulp_policy_t *policy);
SWITCH_DECLARE(void) ulp_increment_successful_partial_recovery_n(ulp_policy_t *policy);

/** Return the number of packets in the buffer. */
SWITCH_DECLARE(uint8_t) ulp_circ_len(ulp_circ_t *circ);

/** Return entry but do not move getp. */
SWITCH_DECLARE(ulp_rtp_msg_t*) ulp_circ_get_soft(ulp_circ_t *circ);

/** Return entry at get_soft index plus offset. */
SWITCH_DECLARE(ulp_rtp_msg_t*) ulp_circ_get_soft_offset(ulp_circ_t *circ, uint8_t offset);

/** Return entry's RTP and move getp. Caller must free it. */
SWITCH_DECLARE(ulp_rtp_pkt_t*) ulp_circ_get(ulp_circ_t *circ);

/** Add to the circular buffer and move putp.
 *  The return value for ULP sender is the number of FEC which can be generated.
 *  The return value for ULP receiver is the max number of protection levels
 *  to which lost packet may be recovered.
 *  Buffer owns the packets and may free them if buffer wraps. */
SWITCH_DECLARE(void) ulp_circ_put_rtp(ulp_policy_t *policy, ulp_rtp_pkt_t *rtp_msg, uint16_t payload_len);

SWITCH_DECLARE(void) ulp_circ_put_fec(ulp_policy_t *policy, ulp_fec_pkt_t *fec, uint16_t payload_len);

/** Discard all packets, free memory, move getp to putp. */
SWITCH_DECLARE(void) ulp_circ_reset(ulp_circ_t *circ);
SWITCH_DECLARE(void) ulp_circ_reset_rtp(ulp_policy_t *policy);
SWITCH_DECLARE(void) ulp_circ_reset_fec(ulp_policy_t *policy);

/** Reset getp_soft to getp. */
SWITCH_DECLARE(void) ulp_circ_reset_soft(ulp_circ_t *circ);

SWITCH_DECLARE(uint8_t) ulp_lvl_n(ulp_policy_t *policy);

/** Set number of levels for this @policy. */
SWITCH_DECLARE(ulp_status_t) ulp_set_lvl_n(ulp_policy_t *policy, uint8_t n);

SWITCH_DECLARE(uint8_t) ulp_lvl_group_size(ulp_policy_t *policy, uint8_t k);

/** Set @policy level's @k group size to @g_k packets. */
SWITCH_DECLARE(ulp_status_t) ulp_set_lvl_group_size(ulp_policy_t *policy, uint8_t k, uint8_t g_k);

SWITCH_DECLARE(uint16_t) ulp_lvl_protection_len(ulp_policy_t *policy, uint8_t k);

/** Set @policy level's @k protection length to @len octets. */
SWITCH_DECLARE(ulp_status_t) ulp_set_lvl_protection_len(ulp_policy_t *policy, uint8_t k, uint16_t len);

SWITCH_DECLARE(uint16_t) ulp_lvl_protection_mask16(ulp_policy_t *policy, uint8_t k);

/** Set @policy level's @k short protection mask to @mask16. */
SWITCH_DECLARE(ulp_status_t) ulp_set_lvl_protection_mask16(ulp_policy_t *policy, uint8_t k, uint16_t mask16);

SWITCH_DECLARE(uint32_t) ulp_lvl_protection_mask_cont(ulp_policy_t *policy, uint8_t k);

/** Set @policy level's @k long protection mask to @mask_cont. */
SWITCH_DECLARE(ulp_status_t) ulp_set_lvl_protection_mask_cont(ulp_policy_t *policy, uint8_t k, uint32_t mask_cont);

/** Return a pointer to the payload of the protection level @k in the FEC packet @fec,
 * or NULL if the FEC packet doesn't contain level @k protection. */ 
SWITCH_DECLARE(uint8_t*) ulp_fec_payload(ulp_fec_lvl_64_info_t *lvl, uint8_t lvl_n, ulp_fec_pkt_t *fec, uint8_t k);

/* Record the total length of the most recent FEC packet (including header). */
SWITCH_DECLARE(void) ulp_fec_set_last_len(ulp_policy_t *policy, uint16_t fec_len);

/* Get the total length of the most recently generated FEC packet (including header). */
SWITCH_DECLARE(uint16_t) ulp_fec_last_len(ulp_policy_t *policy);

/* Return FEC packet length of up to @lvl_n levels. */ 
SWITCH_DECLARE(uint16_t) ulp_fec_max_len(ulp_fec_pkt_t *fec, uint16_t fec_payload_len, uint8_t lvl_n);

/* Record the total length of the most recently recovered RTP packet (including header).
 * This is length of the original packet. */
SWITCH_DECLARE(void) ulp_rtp_set_last_len(ulp_policy_t *policy, uint16_t rtp_len);

/* Get the total length of the most recently recovered RTP (including header). */
SWITCH_DECLARE(uint16_t) ulp_rtp_last_len(ulp_policy_t *policy);

/* Record the recovery length of the most recently recovered RTP packet (including header).
 * This is length of the actually recovered packet (in case of partial recovery it is less than original length). */
SWITCH_DECLARE(void) ulp_rtp_set_last_recover_len(ulp_policy_t *policy, uint16_t rtp_len);

/* Get the recovery length of the most recently recovered RTP (including header). */
SWITCH_DECLARE(uint16_t) ulp_rtp_last_recover_len(ulp_policy_t *policy);

/** Add RTP packet to the ULP buffer. */
SWITCH_DECLARE(void) ulp_add_rtp(ulp_policy_t *policy, ulp_rtp_pkt_t *rtp_msg);

SWITCH_DECLARE(void) ulp_xor(void *x, void *y, size_t len);

/**
 * Parse protection mask, return the group size of the level.
 */
SWITCH_DECLARE(uint8_t) ulp_fec_mask_to_gk(uint64_t mask);

/**
 * Return the @i-th sequence set in the mask. Assume MSB in mask denotes @snb
 * and sequence for the first bit set in the mask is returned for @i == 1.
 */
SWITCH_DECLARE(uint16_t) ulp_mask_2_seq(uint16_t mask16, uint32_t mask_cont, uint16_t snb, uint8_t i);

/**
 * Generate FEC packet for all @fec_n levels of protection: FEC header + FEC levels(header + payload).
 * @fec_n must be less or equal max number of levels that can be generated at the moment,
 * otherwise NULL is returned.
 * If packet is returned, it's length can be calculated from @fec_n and FEC internal state information,
 * or retrieved by call to ulp_fec_last_len().
 *
 * Assume RTP packets with sequences incremented by 1 are used to generate FEC,
 * i.e. packets with sequences in range [fec->hdr.snb, fec->hdr.snb + G(k)) are associated
 * with the FEC.
 */
SWITCH_DECLARE(ulp_fec_pkt_t*) ulp_gen_fec(ulp_policy_t *policy, uint8_t fec_n, uint16_t snb);

/**
 * Return 1 if FEC with @snb protects @seq on level of mask @mask and @mask_cont.
 */
SWITCH_DECLARE(uint8_t) ulp_seq_in_mask(uint16_t snb, uint16_t seq, uint16_t mask16, uint32_t mask_cont);

/**
 * Return 1 if the FEC packet protects @seq on the protection level @k, otherwise return 0.
 * This function parses FEC packet and checks if it contains required level of protection,
 * and if protection mask of that level includes @seq.
 *
 * @seq in host byte order.
 */
SWITCH_DECLARE(uint8_t) ulp_seq_protected_by_fec_on_lvl(ulp_policy_t *policy, ulp_fec_pkt_t *fec, uint16_t fec_payload_len, uint16_t seq, uint8_t k);

/**
 * Return 1 if FEC generated for @snb on max level of protection @k_max would protect @seq on level @k.
 * @snb and @seq in host byte order.
 *
 * Assume n = 2, L0_gk = 2, L1_gk = 4, RTP: 8, 9, 10, 11.
 * Then FEC with snb == 8 protects 10 and 11 on L0 and 8, 9, 10, 11 on L1.
 */
SWITCH_DECLARE(uint8_t) ulp_seq_protected_on_lvl(ulp_policy_t *policy, uint16_t snb, uint8_t k_max, uint16_t seq, uint8_t k);

/**
 * Return the number of RTP packets associated with the @fec packet on level @k and present in the RTP buffer.
 * Packets are put into @rtp table, which must be of size ULP_LVL_GROUP_SIZE_MAX.
 * If the @pad > 0 then pad with 0s all returned packets shorter then @pad octets up to that length,
 * and record new size in the @payload_len (original packet length is always accessible in @payload_original_len).
 * This padding is needed by ULP FEC receiver to mirror ULP FEC sender's protection generation.
 * 
 * Assume RTP packets with sequences incremented by 1 are used to generate FEC,
 * i.e. packets with sequences in range [fec->hdr.snb, fec->hdr.snb + G(k)) are associated
 * with the FEC.
 *
 * Reallocs packets in the circular buffer, saving pointers to these copies
 * in the @rtp table.
 */
SWITCH_DECLARE(uint8_t) ulp_rtp_cover(ulp_policy_t *policy, ulp_fec_pkt_t *fec, uint16_t fec_payload_len, uint8_t k, ulp_rtp_msg_t *rtp, uint16_t pad);

/**
 * Return the number of RTP packets associated with potential FEC with @snb on level @k and present in the RTP buffer.
 * Packets are put into @rtp table, which must be of size ULP_LVL_GROUP_SIZE_MAX.
 * If the @pad > 0 then pad with 0s all returned packets shorter then @pad octets up to that length,
 * and record new size in the @payload_len (original packet length is always accessible in @payload_original_len).
 */
SWITCH_DECLARE(uint8_t) ulp_rtp_material(ulp_policy_t *policy, uint16_t snb, uint8_t k, ulp_rtp_msg_t *rtp, uint16_t pad);

/**
 * Pad RTP packets in @rtp table which areshorter than @pad octets to that length.
 */
SWITCH_DECLARE(ulp_status_t) ulp_rtp_pad(ulp_rtp_msg_t *rtp, uint8_t pkts, uint16_t pad);

/**
 * Free @pkts RTP packets contained in the @rtp table starting from beginning.
 */
SWITCH_DECLARE(void) ulp_rtp_free(ulp_rtp_msg_t *rtp, uint8_t pkts);

/**
 * Return the number of protection levels supported by this FEC packet.
 * 0 indicates that FEC packet doesn't correspond to this @policy,
 * it's length is not aligned with the protection lengths of the levels
 * configured for @policy.
 */
SWITCH_DECLARE(uint8_t) ulp_fec_lvl_n(ulp_policy_t *policy, ulp_fec_pkt_t *fec, uint16_t payload_len);

/**
 * Return the number of FEC packets which protect @seq on level @k.
 * Packets are put into @fecs table, which must be of size ULP_FEC_MAX_N.
 */
SWITCH_DECLARE(uint8_t) ulp_find_fecs_for_seq_on_lvl(ulp_policy_t *policy, uint16_t seq, uint8_t k, ulp_rtp_msg_t *fecs);

/**
 * Return copy of the RTP packet with sequence @seq if it is present in the RTP buf, NULL otherwise.
 * If the RTP packet is returned then @*payload_original_len is set to it's length.
 */
SWITCH_DECLARE(ulp_rtp_pkt_t*) ulp_rtp_by_seq(ulp_policy_t *policy, uint16_t seq, uint16_t *payload_original_len);

#define ULP_RECOVERED_RTP_MAX_N 20

typedef struct ulp_recovered_rtp_s {
	ulp_rtp_pkt_t	*rtp;				/* Recovered versions of the missing RTP packet. */
	uint16_t		rtp_len;			/* Recovered length of the packet. */
	uint8_t			valid;				/* 1 if packet recovery considered valid by validation callbackpassed to @ulp_recover. */
} ulp_recovered_rtp_t;

/**
 * Recover RTP header. For ULP this means recovery on level @k == 0, for non-ULP this happens on any level.
 */
SWITCH_DECLARE(ulp_rtp_pkt_t*) ulp_recover_hdr(ulp_fec_pkt_t *fec, uint16_t seq, uint8_t gk, ulp_rtp_msg_t *rtp, uint16_t *len_r);

/**
 * Recovery procedure. Recover RTP packet with sequence @seq.
 * Return the recovered RTP packet or NULL if recovery was not possible.
 * If only partial recovery was possible @partial is set to 1 on the return,
 * otherwise @*partial is set to 0 indicating complete recovery.
 */
SWITCH_DECLARE(ulp_rtp_pkt_t*) ulp_recover(ulp_policy_t *policy, uint16_t seq, ulp_status_t *err, uint8_t *packet_valid, uint8_t(*cb_validate_recovery)(ulp_rtp_pkt_t *new_rtp, ulp_rtp_msg_t *rtp, uint32_t rtp_n));

SWITCH_DECLARE(ulp_status_t) ulp_encode(ulp_policy_t *policy, void *rtp_hdr, int *len_ptr);

SWITCH_DECLARE(ulp_status_t) ulp_decode(ulp_policy_t *policy, void *rtp_hdr, int *len_ptr);

SWITCH_DECLARE(const char*) ulp_get_version_string(void);

SWITCH_DECLARE(unsigned int) ulp_get_version(void);

SWITCH_DECLARE(void*) ulp_get_user_data(ulp_policy_t *policy);
SWITCH_DECLARE(void) ulp_set_user_data(ulp_policy_t *policy, void* user_data);


/**
 * RED
 *
 *
 * Note: supports single Redundant Encoding Block.
 * TODO: add support for multiple blocks of redundant encoding.
 *
 * Packet has to carry a primary encoding and one or more redundant encodings. As the use of variable size encodings is desirable,
 * each encoded block in the packet has to have a length indicator.
 *
 * +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
 * |                 RTP Header (RED) - 12 octets                  | Timestamp, ssrc and seq are same as Primary Encoding's timestamp.
 * +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
 * |       Redundant Encoding Block Header (RED) - 4 octets        |
 * +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
 * |                        FEC Packet Data                        |
 * +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
 * |        Primary Encoding Block Header (RED) - 1 octet          | 4 or 1 octets. May be just 1 octet as the ts of Primary Encoding is same as that of the RTP header and length can be determined from total packet length.
 * +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
 * |                       Media Packet Data                       |
 * +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
 *
 *
 * RTP Header
 *
 * The RTP header provides a timestamp field that corresponds to
 * the time of creation of the encoded data. When redundant
 * encodings are used this timestamp field can refer to the time of creation
 * of the primary encoding data. Redundant blocks of data will correspond
 * to different time intervals than the primary data, and hence each block
 * of redundant encoding will require its own timestamp. To reduce the number
 * of bytes needed to carry the timestamp, it can be encoded as the difference
 * of the timestamp for the redundant encoding and the timestamp of the primary.
 *
 * Note:
 *
 * RFC 5109: 10.3
 * In the payload format for redundant coding as specified by RFC 2198,
 * the marker bit is lost as soon as the primary coding is carried in the RED packets.
 * So the marker bit cannot be recovered regardless of whether or not the FEC is used.
 *
 * RFC 2198: 4  Limitations
 * The RTP marker bit is not preserved for redundant data blocks. Hence if the primary
 * (containing this marker) is lost, the marker is lost.
 *
 *
 * Encoding Header(s)
 * 0                   1                    2                   3
 * 0 1 2 3 4 5 6 7 8 9 0 1 2 3  4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
 * +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
 * |F|   block PT  |  timestamp offset         |   block length    |
 * +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
 *
 *
 * Primary Encoding Block Header
 *
 * The primary encoding block header is placed last in the packet. It is therefore possible to omit the timestamp and block-length fields
 * from the header of this block, since they may be determined from the RTP header and overall packet length. The header for the primary (final) block
 * comprises only a zero F bit, and the block payload type information, a total of 8 bits. This header is followed by RTP payload (less header) of the media.
 *
 *
 * Redundant Encoding Block Header
 *
 * Contains PT=127 for FEC. Followed by FEC (FEC Header + for all ULP levels (header + payload [protection]).
 *
 *
 * NOTE:
 *
 * Primary Encoding doesn't include RTP header, as it is copied to RED header, which differs only in PT (and Marker field is lost).
 * So the media is extracted from RED by copying data following Primary Encoding Block Header and prepending
 * it with RTP header of the RED, changing PT to that of primary's encoding.
 *
 * RFC 5109 10.3:
 * The FEC data is generated not directly from the RED packets, but from the virtual RTP packets
 * containing the media packet data. Those virtual RTP packets can be very easily generated
 * from the RED packets both with and without redundant coding included. The conversion from RED packets
 * to virtual RTP packets is simply done by (1) removing any RED block headers and redundant coding data,
 * and (2) replacing the PT in the RTP header with the PT of the primary coding.
*/


/**
 * Generate RTP header for RED.
 * Return pointer to RED packet or NULL on failure.
 * Timestamp, ssrc and seq are copied from @rtp header.
 * If @inline_where is NULL, then allocate memory and return pointer to it, otherwise
 * generate data in the place pointed to by @inline_where (must be >= 12 octets in length).
 *
 * Note:
 *
 * RFC 5109: 10.3
 * In the payload format for redundant coding as specified by RFC 2198,
 * the marker bit is lost as soon as the primary coding is carried in the RED packets.
 * So the marker bit cannot be recovered regardless of whether or not the FEC is used.
 *
 * RFC 2198: 4  Limitations
 * The RTP marker bit is not preserved for redundant data blocks. Hence if the primary
 * (containing this marker) is lost, the marker is lost.
 */
SWITCH_DECLARE(ulp_rtp_hdr_t*) ulp_gen_red_hdr(ulp_rtp_hdr_t *rtp, uint8_t red_pt, void *inline_where);

/**
 * Return the length of RED payload (encodings + their payloads), i.e length of RED packet less header.
 * @rtp_len - length of the RTP payload (RTP packet excluding header)
 * @redundancy_len - total length of the Redundant Encoding, in case of ULPFEC this is total FEC packet length (including FEC Header)
 */
SWITCH_DECLARE(uint16_t) ulp_red_data_len(uint16_t rtp_len, uint16_t redundancy_len);

/**
 * Generate RED packet payload: Primary Encoding Block Header and payload (media), plus optional Redundant Encoding Block Header and payload (FEC).
 *
 * Return pointer to RED payload or NULL on failure.
 * If @inline_where is NULL, then allocate memory and return pointer to it, otherwise
 * generate data in the place pointed to by @inline_where (must be >= ulp_red_data_len() in length).
 *
 * @rtp - points to media (including RTP header)
 * @rtp_len - length of the RTP payload (RTP packet excluding header)
 * @primary_pt - payload type of Primary Encoding, should be same as the payload of the transmitted RTP media packet
 * @fec - points to FEC (total packet including FEC Header + all levels (header + payload))
 * @fec_len - length of the FEC payload (FEC packet excluding header)
 * @total_data_len - (out) on successful RED payload generation it is set to total length of the RED payload (excluding RED RTP header),
 *					on error it is set to 0 (though this is NULL returned value that should be checked)
 * @inline_where - if NULL, the memory needed for RED packet payload (excluding RED RTP header) is allocated, otherwise payload is created in memory
 *					pointed to by that (so it must be at least ulp_red_data_len()) 
 *
void* ulp_gen_red_data(ulp_rtp_hdr_t *rtp, uint16_t rtp_len, uint8_t primary_pt, ulp_fec_pkt_t *fec, uint16_t fec_len, uint16_t *total_data_len, void *inline_where);*/

/**
 * Generate RED packet payload: Primary Encoding Block Header and payload (media), plus optional Redundant Encoding Block Header and payload (FEC).
 *
 * Return pointer to RED payload or NULL on failure.
 * If @inline_where is NULL, then allocate memory and return pointer to it, othergenerate data in the place pointed to by @inline_where (must be >= ulp_red_data_len() in length).
 * generate data in the place pointed to by @inline_where (must be >= ulp_red_data_len() in length).
 *
 * @rtp - points to media (including RTP header)
 * @rtp_len - length of the RTP payload (RTP packet excluding header)
 * @primary_pt - payload type of Primary Encoding, should be same as the payload of the transmitted RTP media packet
 * @redundancy - points to redundant data, in case of ULPFEC scheme it is FEC packet (total packet including FEC Header + all levels (header + payload))
 * @redundancy_len - total length of the redundancy (in case of FEC it is length of the packet including FEC header and all protection levels [header + payload])
 * @redundancy_pt - payload of the Redundant Encoding, for ULPFEC it is 127
 * @red_ts_off_diff - TS to be copied into Redundant Encoding ts field
 * @total_data_len - (out) on successful RED payload generation it is set to total length of the RED payload (excluding RED RTP header),
 *					on error it is set to 0 (though this is NULL returned value that should be checked)
 * @inline_where - if NULL, the memory needed for RED packet payload (excluding RED RTP header) is allocated, otherwise payload is created in memory
 *					pointed to by that (so it must be at least ulp_red_data_len()) 
 */
SWITCH_DECLARE(void*) ulp_gen_red_data(ulp_rtp_hdr_t *rtp, uint16_t rtp_len, uint8_t primary_pt, void *redundancy, uint16_t redundancy_len, uint8_t redundancy_pt, uint16_t red_ts_off_diff, uint16_t *total_data_len, void *inline_where);

/**
 * Generate RED packet: RED header + data (encodings + their payloads).
 * Return pointer to RED packet or NULL on failure.
 * @rtp	- points to media (including RTP header)
 * @rtp_len	- length of the RTP payload (RTP packet __excluding__ header)
 * @fec	- points to FEC (total packet including FEC Header + all levels (header + payload))
 * @fec_len - length of the FEC payload (FEC packet __excluding__ header)
 * @total_len - (out) on successful RED generation is set to total length of the RED packet (including RTP header),
 *				on error is set to 0 (though this is NULL returned value what indicates failure)
 * @red_pt - PT field in RTP header of created RED packet is set to that
 * @inline_where - if NULL, all needed memory for RED packet is allocated, otherwise packet is created in memory
 *				pointed to by that (so it must be at least ulp_red_data_len() + sizeof(RTP header)  
 *
void* ulp_gen_red_pkt(ulp_rtp_hdr_t *rtp, uint16_t rtp_len, void *fec, uint16_t fec_len, uint16_t *total_len, uint8_t red_pt, void* inline_where);
*/

/**
 * Generate RED packet: RED header + data (encodings + their payloads).
 *
 * Return pointer to RED packet or NULL on failure.
 *
 * Note:
 *
 * RFC 5109: 10.3
 * In the payload format for redundant coding as specified by RFC 2198,
 * the marker bit is lost as soon as the primary coding is carried in the RED packets.
 * So the marker bit cannot be recovered regardless of whether or not the FEC is used.
 *
 * RFC 2198: 4  Limitations
 * The RTP marker bit is not preserved for redundant data blocks. Hence if the primary
 * (containing this marker) is lost, the marker is lost.
 *
 * @rtp - points to media (including RTP header)
 * @rtp_len - length of the RTP payload (RTP packet __excluding__ header)
 * @redundancy - points to redundant data, in case of ULPFEC scheme it is FEC packet (total packet including FEC Header + all levels (header + payload))
 * @redundancy_len - total length of the redundancy (in case of FEC it is length of the packet including FEC header and all protection levels [header + payload])
 * @redundancy_pt - payload of the Redundant Encoding, for ULPFEC it is 127
 * @red_ts_off_diff - TS to be copied into Redundant Encoding ts field
 * @total_len - (out) on successful RED generation is set to total length of the RED packet (including RTP header),
 *				on error is set to 0 (though this is NULL returned value what indicates failure)
 * @red_pt - PT field in RTP header of created RED packet
 * @inline_where - if NULL, all needed memory for RED packet is allocated, otherwise packet is created in memory
 * pointed to by that (so it must be at least ulp_red_data_len() + sizeof(RTP header)
 */
SWITCH_DECLARE(void*) ulp_gen_red_pkt(ulp_rtp_hdr_t *rtp, uint16_t rtp_len, void *redundancy, uint16_t redundancy_len, uint8_t redundancy_pt, uint16_t red_ts_off_diff, uint16_t *total_len, uint8_t red_pt, void* inline_where);

/**
 * Depaketize RED packet into Primary and Redundant encodings.
 *
 * Return pointer to the Primary Encoding (media), on failure return NULL.
 *
 * Note:
 *
 * RFC 5109: 10.3
 * In the payload format for redundant coding as specified by RFC 2198,
 * the marker bit is lost as soon as the primary coding is carried in the RED packets.
 * So the marker bit cannot be recovered regardless of whether or not the FEC is used.
 *
 * RFC 2198: 4  Limitations
 * The RTP marker bit is not preserved for redundant data blocks. Hence if the primary
 * (containing this marker) is lost, the marker is lost.
 *
 * @rtp - (in) should point to RED packet (including header),
 *			on return @rtp is set to unpacked Primary Encoding. If @inline_where is NULL, then Primary Encoding
 *			is copied to new malloced memory, otherwise Primary Encoding is copied to memory pointed to by @inline_where
 * @rtp_len - (in/out) on the call to this method @rtp_len should be set to length of the RED payload (excluding header),
 *				on return @rtp_len is set to the length of the extracted Primary Encoding (excluding header)
 * @redundancy - (out) on the return points to the extracted Redundant Encoding if any
 *					(NULL if there was no Redundant Encoding)
 * @redundancy_len - (out) on the return points to the total length of the Redundant Encoding (retrieved from Redundant Encoding Block Header),
 *						zero if there was no Redundant Encoding
 * @red_ts_off_diff - (out) timestamp offset field from Redundant Encoding Block Header (redundant encoding's timestamp offset from Primary Encoding),
 *						zero if there was no Redundant Encoding
 * @inline_where - configures where to extract Primary Encoding. If @inline_where is NULL, then Primary Encoding
 *					is copied to the new malloced memory, otherwise Primary Encoding is copied to the memory pointed to by @inline_where
 *					(so it can't overlap with @rtp and must be long enough, if it is not shorter than original RED packet, then this will work Okay) 
 */
SWITCH_DECLARE(ulp_rtp_hdr_t*) ulp_red_depaketize(ulp_rtp_hdr_t *rtp, uint16_t *rtp_len, void **redundancy, uint16_t *redundancy_len, uint16_t *red_ts_off_diff, void* inline_where);

/**
 * Understand FEC setup for the given @fec packet.
 *
 * Return number of protection levels present in @fec packet and all level headers info, up to @lvl_n, in host byte order.
 *
 * @fec - FEC packet to parse (policy to understand)
 * @fec_len - (in) total length of the FEC packet (header + payload)
 * @lvl - (in/out) array of level header info filled in by this function
 * @lvl_n - (in/out) on entry indicates the number of entries in @lvl table, on return indicates the number of levels found in FEC and filled in the array.
 *			If @more is 1 then there was more levels than entries in the array.
 * @more - (out) indicates if there is more levels in this FEC than @lvl_n on entry
 * @l - (out) indicates if FEC level headers use long mask
 */
SWITCH_DECLARE(ulp_status_t) ulp_fec_parse(ulp_policy_t *policy, ulp_fec_pkt_t *fec, uint16_t fec_len, ulp_fec_lvl_64_info_t *lvl, uint8_t *lvl_n,  uint8_t *more, uint8_t *l);

#define ulp_rtp_fec_2_idx(rtp_n, fec_n) (((rtp_n) * ((rtp_n) - 1) / 2) + (fec_n) - 1)
#define ulp_binomial_table_n_k_2_idx(n, k) (((1 + (n)) * (n) / 2) + (k))

extern uint16_t* ulp_masks16[ULP_MASK16_TABLE_SIZE];
extern uint16_t* ulp_burst_high_loss_masks16[ULP_MASK16_RAND_LOSS_TABLE_SIZE];
extern uint16_t* ulp_burst_medium_loss_masks16[ULP_MASK16_RAND_LOSS_TABLE_SIZE];
extern uint16_t* ulp_burst_small_loss_masks16[ULP_MASK16_RAND_LOSS_TABLE_SIZE];
extern uint16_t** ulp_burst_loss_masks_sets_list[ULP_MAX_MASK_SETS];
extern mask_set_t* ulp_burst_loss_masks_sets_list_set[ULP_MAX_MASK_SETS];

extern uint16_t* ulp_rand_high_loss_masks16[ULP_MASK16_RAND_LOSS_TABLE_SIZE];
extern uint16_t* ulp_rand_medium_loss_masks16[ULP_MASK16_RAND_LOSS_TABLE_SIZE];
extern uint16_t* ulp_rand_small_loss_masks16[ULP_MASK16_RAND_LOSS_TABLE_SIZE];
extern uint16_t** ulp_rand_loss_masks_sets_list[ULP_MAX_MASK_SETS];
extern mask_set_t* ulp_rand_loss_masks_sets_list_set[ULP_MAX_MASK_SETS];
extern uint16_t*** ulp_loss_masks_sets_list;
extern  mask_set_t** ulp_loss_masks_sets_list_set;

SWITCH_DECLARE(void) ulp_set_scheme_masks_sets_type(int loss_type);
SWITCH_DECLARE(int) ulp_get_scheme_fec_count(int set_idx, int mask_idx);
SWITCH_DECLARE(int) ulp_get_scheme_media_count(int set_idx, int mask_idx);

extern uint32_t ulp_binomial_table[(1 + ULP_MAX_FEC_IN_SCHEME) * (1 + ULP_MAX_FEC_IN_SCHEME + 1) / 2];

/**
 * Return binomial coefficient by recursive call.
 */
SWITCH_DECLARE(uint32_t) ulp_binomial(uint8_t n, uint8_t k);

/**
 * Return the optimum probability of successful recovery for the FEC scheme
 * which maximizes probability of successful recovery giving current loss rate
 * of @loss_per_255 packets per 255 sent. Set @fec_n and @rtp_n to FEC number
 * and RTP media number of packets in chosen scheme.
 *
 * This works by searching for best probability over all FEC schemes up to ULP_MAX_FEC_IN_SCHEME
 * number of FECs in scheme.
 *
 * @max_overhead - if not 0 then it is a maximum allowed overhead for the FEC scheme,
 * expressed as a fraction of 255 packets (so 10% overhead is 25).
 */
SWITCH_DECLARE(double) ulp_opt_fec_scheme(uint8_t loss_per_255, uint8_t max_overhead_per_255, uint8_t *fec_n, uint8_t *rtp_n);

/* For testing. */

SWITCH_DECLARE(ulp_rtp_pkt_t*) ulp_gen_rtp(uint8_t v, uint8_t p, uint8_t x, uint8_t cc, uint8_t m, uint8_t pt, uint16_t seq, uint32_t ts, uint32_t ssrc, uint16_t payload_len);
SWITCH_DECLARE(ulp_rtp_pkt_t*) ulp_gen_rtp_host(uint8_t v, uint8_t p, uint8_t x, uint8_t cc, uint8_t m, uint8_t pt, uint16_t seq, uint32_t ts, uint32_t ssrc, uint16_t payload_len);

SWITCH_DECLARE(int) ulp_test_bytes(char *x, char *y, int n);
SWITCH_DECLARE(void) ulp_die(const char *s, const char *file, int line);

#define ulp_circ_inc(p) p += 1; \
						p %= ULP_LVL_GROUP_SIZE_MAX;

#define ulp_test(x, m) if (!(x)) ulp_die(m, __FILE__, __LINE__);

#define ulp_min(x, y) ((x) < (y) ? (x) : (y))
#define ulp_max(x, y) ((x) < (y) ? (y) : (x))

#define ULP_FL_ALLOW_FEC_COLLISION 1u
#define ULP_FL_RCV 4u
#define ULP_FL_SND 8u
				
/* Chromium sends Redundant Encoding as Primary, and the protection length field counts the ULP FEC Level Header too. */
#define ULP_FL_REDUNDANT_AS_PRIMARY 16u

/* Use uneven levels of protections. Off by default. */
#define ULP_FL_USE_ULP 32u

#define ulp_set_flag(p, f) ((p)->flags |= (f))
#define ulp_clear_flag(p, f) ((p)->flags &= ~(f))
#define ulp_test_flag(p, f) (((p)->flags & (f)) ? 1 : 0)

#ifdef __cplusplus
}
#endif

#endif /* ULP_ULP_H */
